/***********************************************************************/
/*                                                                     */
/*                               Pio (C)                               */
/*                                                                     */
/*                    Polyglot Input/Output Library                    */
/*                                                                     */
/*               EPI Pomdapi, INRIA Paris - Rocquencourt               */
/*                                                                     */
/*  Copyright 2011-2014 INRIA.                                         */
/*  Institut National de Recherche en Informatique et en Automatique.  */
/*  All rights reserved. This file is distributed only by permission.  */
/*                                                                     */
/*  Bruno Verlyck <Bruno.Verlyck@inria.fr>                             */
/*                                                                     */
/***********************************************************************/

#include "common.h"
#include "pio_ibuff.h"

#include <strings.h>

void ibuf_of_string (pio_input_buffer *ib, struct string *str) {
  ib->IB_at_eof = false;
  ib->IB_stream = NULL;
  ib->IB_endpos = str->ssize;
  buf_create (&ib->IB_buf, str->ssize + 1); /* 1 for final NUL */
  /* so we can blindly free (?) and let buf_create blindly allocate */
  memcpy (ib->IB_buf.buf_fer, str->str, str->ssize);
  ib->IB_buf.buf_fer[str->ssize] = '\0';
}

void ibuf_of_stream (pio_input_buffer *ib, FILE *stream) {
  memset (ib, 0x55, sizeof (*ib));
  ib->IB_at_eof = false;
  ib->IB_stream = stream;
  ib->IB_endpos = 0;			/* still not filled */
  size_t sz = 2;			/* XXX 1024 */
  buf_create (&ib->IB_buf, sz);
}

/* XXX a bouffe toujours plus de place */
/* XXX Peut-on enlever eof de pio_input_buffer ? */
bool refill (pio_input_buffer *ib) {
  size_t nread = 0;
  if (ib->IB_stream == NULL) {		/* comes from ibuf_of_string */
    ib->IB_at_eof = true;
  } else {
    const size_t pos = ib->IB_endpos;
    {
      /* we want to ensure from endpos, but ensure only knows about buf_position */
      const size_t sv_pos = ib->IB_buf.buf_position;
      ib->IB_buf.buf_position = pos;
      (void) ensure(&ib->IB_buf, sizeof (char) + 1); /* + NUL */
      ib->IB_buf.buf_position = sv_pos;
    }
    const size_t avail = ib->IB_buf.buf_length - 1 - pos; /* + NUL */
    assert (avail > 0);
    nread = fread (&ib->IB_buf.buf_fer[pos], 1, avail, ib->IB_stream);
    if (nread == 0)
      ib->IB_at_eof = true;
    ib->IB_endpos += nread;
    ib->IB_buf.buf_fer[ib->IB_endpos] = '\0';
  }
  return ib->IB_at_eof;
}

static void bad_escape (const char *s) __attribute__ ((noreturn));

static void bad_escape (const char *s) {
  failwithf (__FILE__ "Illegal backslash escaped character (\\%.3s)", s);
}

#define failwith(msg) failwith (__FILE__ "::" msg)

jmp_buf *eof_bscanf_label = NULL;
/* handler to externally decide if scanning is ok after all */
bool (*on_unknown_scan) (pio_input_buffer *);

#define funname "ibuf_scanf"
#define nc_scanned_unset (-27)
/* avoid reading one extra char when at end of a constant format (= ntok_expected == 0 %n);
   else makes some master/slave test fail */
#define REFILL(ib) ((ntok_expected == 0 && (size_t) ib->IB_nc_scanned >= (strlen (format) - 2)) || refill (ib))

void __ibuf_scanf (pio_input_buffer *ib, const size_t ntok_expected, const char *format, ...) {
  va_list args;
  const size_t pos = ib->IB_buf.buf_position;
  int prev_nc_scanned = -1;

  while (true) {
    va_start (args, format);
    ib->IB_nc_scanned = nc_scanned_unset;	/* marker to see if %n was executed */
    const size_t ntok_read =
      vsscanf (&ib->IB_buf.buf_fer[pos], format, args);

    bool nc_scanned_unknown =			/* if true, ntok_read is undefined */
      ib->IB_nc_scanned == nc_scanned_unset;
    if (!nc_scanned_unknown && ntok_read > ntok_expected)
      assert (0);				/* ntok_expected too small at call site */
    else if (nc_scanned_unknown ||		/* 2 kinds of matching failure */
	     ntok_read < ntok_expected) {
      if (nc_scanned_unknown &&			/* apply handler only to 1st kind */
	  on_unknown_scan &&
	  (ib->IB_nc_scanned = 0,		/* don't move buf_position */
	   on_unknown_scan (ib))) {		/* caller decides it's not a failure */
	break;
      }
      if (refill (ib)) {			/* if it's a real failure, we read until eof */
						/* XXX but what else ? limit to X ko ? */
	if (eof_bscanf_label)			/* XXX matchfail_bscanf_label first ? */
	  longjmp (*eof_bscanf_label, (int)((long)&eof_bscanf_label));
	else
	  failwith ("eof uncatched in ibuf_scanf");
      }
    } else {
      /* ntok_read OK but we could nevertheless have to refill, in case
	 the end of the last token is still unread. */
      size_t avail = ib->IB_endpos - pos;
      if ((unsigned) ib->IB_nc_scanned < avail)
	/* success: scan didn't go upto endpos */
	break;
      /* avoid endless loop if we scanned the same number of chars twice in a row */
      if (ib->IB_nc_scanned == prev_nc_scanned)
	break;
      prev_nc_scanned = ib->IB_nc_scanned;
      /* Here we scanned upto or past endpos... */
      if (REFILL (ib)) {			/* eof */
	if ((unsigned) ib->IB_nc_scanned == avail) /* last token ends at eof: success */
	  break;
	failwith (funname ": eof but sscanf went past endpos!");
      }
    }
  }
  ib->IB_buf.buf_position += ib->IB_nc_scanned;
}
#undef funname

char ibuf_getchar (pio_input_buffer *ib) {
  /* We sometimes must avoid ibuf_scanf (ib, 1, "%c", &c) that read one more char than necessary.
     This blocks some master/slave tests. */
  size_t pos = ib->IB_buf.buf_position;
  if ((ib->IB_endpos - pos) == 0)
    refill (ib);
  char c = ib->IB_buf.buf_fer[pos];
  ib->IB_buf.buf_position = ++pos;
  return c;
}

static char read_escaped_char (pio_input_buffer *ib) {
  char esc_seq[6];
  bzero (&esc_seq[0], sizeof (esc_seq));
  ibuf_scanf (ib, 1, "%c", &esc_seq[0]);
  bool hex = false;
  switch (esc_seq[0]) {
  case '\\':
  case '\"':
  case '\'':
  case ' ':  return esc_seq[0];
  case 'n':  return '\n';
  case 't':  return '\t';
  case 'r':  return '\r';
  case 'b':  return '\b';
  case 'x':
    hex = true;				/* fall through */
  case '0': case '1': case '2': case '3': case '4':
  case '5': case '6': case '7': case '8': case '9':
    if (ib->IB_stream) {
      /* simpler to first read it entirely */
      ensure (&ib->IB_buf, 3);
      refill (ib);
    }
    ibuf_scanf (ib, 1, hex ? "%2[0-9a-fA-F]" : "%2[0-9]", &esc_seq[1]);
    if (ib->IB_nc_scanned != 2)
      bad_escape (&esc_seq[0]);
    return (char) strtol (&esc_seq[hex], NULL, hex ? 16 : 10);
  default:
    bad_escape (&esc_seq[0]);
  }
}

static bool rds_handler (pio_input_buffer *ib) {
  char c = ib->IB_buf.buf_fer[ib->IB_buf.buf_position];
  return (c == '\"' || c == '\\');
}

struct string read_delimited_string (pio_input_buffer *ib, size_t initial) {
  bool (*sv_on_unknown_scan) (pio_input_buffer *) = on_unknown_scan;
  pio_buffer ob;

  buf_create (&ob, initial);
  ibuf_scanf0 (ib, "\"");
  bool done = false;
  while (!done) {
    (void) ensure (&ob, sizeof (char) + 1);	/* we need to be able to store 1 char + final NUL at least */
    char format[30];
    size_t ppos = ob.buf_position;
    /* XXX att: %[] veut au moins 1 char en C, or il peut y en avoir 0 */
    snprintf (format, sizeof (format) - 1, "%%%lu[^\\\"]%%n", (long unsigned int) (ob.buf_length - 1 - ppos)); /* - 1 : final NUL */ 
    on_unknown_scan = rds_handler;
    __ibuf_scanf (ib, 1, format, &ob.buf_fer[ppos], &ib->IB_nc_scanned);
    ob.buf_position += ib->IB_nc_scanned;
    if (true || ib->IB_nc_scanned == 0) {
      /* at eof, " or \ */
      char delim = ibuf_getchar (ib);
      /* XXX vrifier que sur eof, on sort bien de ibuf_scanf par exception */
      switch (delim) {
      case '\"':
	done = true;			/* otherwise would need break 2 */
	break;
      case '\\':
	buf_add_char (&ob, read_escaped_char (ib));
	break;
      default:
	/* initial too short */
	buf_add_char (&ob, delim);
      }
    }
  }
  on_unknown_scan = sv_on_unknown_scan;		/* XXX but non-local jumps... */
  struct string str = buf_contents (&ob);
  free (ob.buf_fer);
  return str;
}
